from chameleon.core import config

import chameleon.zpt.language
import pagetemplate

class XHTMLElement(chameleon.zpt.language.XHTMLElement):
    meta_interpolation_escaping = False

class Parser(chameleon.zpt.language.Parser):
    default_expression = "path"

    element_mapping = {
        config.META_NS: {None: XHTMLElement},
        }
    
class TextTemplate(pagetemplate.PageTemplate):
    __doc__ = pagetemplate.PageTemplate.__doc__ # for Sphinx autodoc
    default_parser = Parser()
    format = 'text'

class TextTemplateFile(pagetemplate.PageTemplateFile):
    __doc__ = pagetemplate.PageTemplateFile.__doc__ # for Sphinx autodoc
    default_parser = Parser()
    format = 'text'

class ViewTextTemplate(pagetemplate.ViewPageTemplate):
    __doc__ = pagetemplate.ViewPageTemplate.__doc__ # for Sphinx autodoc
    default_parser = Parser()
    format = 'text'

class ViewTextTemplateFile(pagetemplate.ViewPageTemplateFile):
    __doc__ = pagetemplate.ViewPageTemplateFile.__doc__ # for Sphinx autodoc
    default_parser = Parser()
    format = 'text'



import re
import namespaces
import zope.event

from zope.traversing.adapters import traversePathElement
from zope.contentprovider.interfaces import IContentProvider
from zope.contentprovider.interfaces import ContentProviderLookupError
from zope.traversing.interfaces import ITraversable

try:
    from zope.contentprovider.interfaces import BeforeUpdateEvent
except ImportError:
    BeforeUpdateEvent = None

from chameleon.core import types
from chameleon.zpt import expressions
from chameleon.zpt.interfaces import IExpressionTranslator

from types import MethodType

_marker = object()
_valid_name = re.compile(r"[a-zA-Z][a-zA-Z0-9_]*$").match

def identity(x):
    return x

class ContentProviderTraverser(object):
    def __call__(self, context, request, view, name):
        cp = zope.component.queryMultiAdapter(
            (context, request, view), IContentProvider, name=name)

        # provide a useful error message, if the provider was not found.
        if cp is None:
            raise ContentProviderLookupError(name)

        if BeforeUpdateEvent is not None:
            zope.event.notify(BeforeUpdateEvent(cp, request))
        cp.update()
        return cp.render()

class ZopeTraverser(object):
    def __init__(self, proxify=identity):
        self.proxify = proxify

    def __call__(self, base, request, call, *path_items):
        """See ``zope.app.pagetemplate.engine``."""

        if bool(path_items):
            path_items = list(path_items)
            path_items.reverse()

            while len(path_items):
                name = path_items.pop()
                ns = ':' in name
                if ns is True:
                    namespace, name = name.split(':', 1)
                    base = namespaces.function_namespaces[namespace](base)
                    if ITraversable.providedBy(base):
                        base = self.proxify(traversePathElement(
                            base, name, path_items, request=request))
                        continue

                # special-case dicts for performance reasons
                if isinstance(base, dict):
                    next = base.get(name, _marker)
                else:
                    next = getattr(base, name, _marker)

                if next is not _marker:
                    base = next
                    if ns is True and isinstance(base, MethodType):
                        base = base()
                    continue
                else:
                    base = traversePathElement(
                        base, name, path_items, request=request)

                if not isinstance(base, (basestring, tuple, list)):
                    base = self.proxify(base)

        if call and getattr(base, '__call__', _marker) is not _marker:
            return base()

        return base

class ZopeExistsTraverser(ZopeTraverser):
    exceptions = AttributeError, LookupError, TypeError

    def __call__(self, base, request, call, *args, **kwargs):
        try:
            return ZopeTraverser.__call__(
                self, base, request, False, *args, **kwargs) is not None
        except self.exceptions:
            return False
        return True

class PathTranslator(expressions.ExpressionTranslator):
    path_regex = re.compile(
        r'^((nocall|not):\s*)*([A-Za-z_][A-Za-z0-9_:]*)'+
        r'(/[?A-Za-z_@\-+][?A-Za-z0-9_@\-\.+/:]*)*$')

    interpolation_regex = re.compile(
        r'\?[A-Za-z][A-Za-z0-9_]+')

    path_traverse = ZopeTraverser()
    scope = 'request'

    symbol = '_path'

    def translate(self, string, escape=None):
        """
        >>> translate = PathTranslator().translate

        >>> translate("") is None
        True

        >>> translate("nocall: a")
        value('a')

        >>> translate("nothing")
        value('None')

        >>> translate("a/b")
        value("_path(a, request, True, 'b')")

        Verify allowed character set.

        >>> translate("image_path/++res++/@@hello.html")
        value("_path(image_path, request, True, '++res++', '@@hello.html')")

        >>> translate("context/@@view")
        value("_path(context, request, True, '@@view')")

        >>> translate("nocall: context/@@view")
        value("_path(context, request, False, '@@view')")

        >>> translate("context/?view")
        value("_path(context, request, True, '%s' % (view,))")

        >>> translate("context/@@?view")
        value("_path(context, request, True, '@@%s' % (view,))")
        """

        if not string:
            return None

        if not self.path_regex.match(string.strip()):
            raise SyntaxError("Not a valid path-expression: %s." % string)

        nocall = False

        while string:
            m = self.re_pragma.match(string)
            if m is None:
                break

            string = string[m.end():]
            pragma = m.group('pragma').lower()

            if pragma == 'nocall':
                nocall = True
            else:
                raise ValueError("Invalid pragma: %s" % pragma)

        parts = string.strip().split('/')

        # map 'nothing' to 'None'
        parts = map(lambda part: part == 'nothing' and 'None' or part, parts)

        components = []
        for part in parts[1:]:
            interpolation_args = []

            def replace(match):
                start, end = match.span()
                interpolation_args.append(
                    part[start+1:end])
                return "%s"

            while True:
                part, count = self.interpolation_regex.subn(replace, part)
                if count == 0:
                    break

            if len(interpolation_args):
                component = "%s %% (%s,)" % (
                    repr(part), ", ".join(interpolation_args))
            else:
                component = repr(str(part))

            components.append(component)

        base = parts[0]

        if not components:
            if len(parts) == 1 and (nocall or base == 'None'):
                value = types.value('%s' % base)
                return value
            else:
                components = ()

        value = types.value(
            '%s(%s, %s, %s, %s)' % \
            (self.symbol, base, self.scope, not nocall, ', '.join(components)))

        value.symbol_mapping[self.symbol] = self.path_traverse

        return value

class NotTranslator(expressions.ExpressionTranslator):
    zope.component.adapts(IExpressionTranslator)

    recursive = True

    def __init__(self, translator):
        self.translator = translator

    def tales(self, string, escape=None):
        """
        >>> tales = NotTranslator(path_translator).tales

        >>> tales("abc/def/ghi")
        value("not(_path(abc, request, True, 'def', 'ghi'))")

        >>> tales("abc | def")
        parts(value('not(_path(abc, request, True, ))'),
              value('not(_path(def, request, True, ))'))

        >>> tales("abc | not: def")
        parts(value('not(_path(abc, request, True, ))'),
              value('not(not(_path(def, request, True, )))'))

        >>> tales("abc | not: def | ghi")
        parts(value('not(_path(abc, request, True, ))'),
              value('not(not(_path(def, request, True, )))'),
              value('not(not(_path(ghi, request, True, )))'))
        """

        value = self.translator.tales(string, escape=escape)
        if isinstance(value, types.value):
            value = (value,)

        parts = []
        for part in value:
            factory = type(part)
            value = factory("not(%s)" % part)
            value.symbol_mapping.update(part.symbol_mapping)
            parts.append(value)

        if len(parts) == 1:
            return parts[0]

        return types.parts(parts)

class ProviderTranslator(expressions.ExpressionTranslator):
    provider_regex = re.compile(r'^[A-Za-z][A-Za-z0-9_\.-]*$')

    symbol = '_get_content_provider'
    content_provider_traverser = ContentProviderTraverser()

    def translate(self, string, escape=None):
        if self.provider_regex.match(string) is None:
            raise SyntaxError(
                "%s is not a valid content provider name." % string)

        value = types.value("%s(context, request, view, '%s')" % \
                            (self.symbol, string))
        value.symbol_mapping[self.symbol] = self.content_provider_traverser
        return value

class ExistsTranslator(PathTranslator):
    """Implements string translation expression."""

    symbol = '_path_exists'

    path_traverse = ZopeExistsTraverser()

    def translate(self, *args, **kwargs):
        value = super(ExistsTranslator, self).translate(*args, **kwargs)
        if value is None:
            return

        assert isinstance(value, types.value)
        parts = types.parts(
            (value, types.value('False')))
        parts.exceptions = NameError,
        return parts

exists_translator = ExistsTranslator()
path_translator = PathTranslator()
provider_translator = ProviderTranslator()



import os
import sys
import compiler

from zope import i18n
from zope import component

from chameleon.core import types
from chameleon.core import config
from chameleon.core import codegen
from chameleon.core import clauses
from chameleon.core import generation
from chameleon.core import utils
from chameleon.core.i18n import fast_translate
from chameleon.zpt import template
from chameleon.zpt.interfaces import IExpressionTranslator

from z3c.pt import language

_marker = object()
_expr_cache = {}

class opaque_dict(dict):
    def __new__(cls, dictionary):
        inst = dict.__new__(cls)
        inst.dictionary = dictionary
        return inst

    @property
    def __getitem__(self):
        return self.dictionary.__getitem__

    @property
    def __len__(self):
        return self.dictionary.__len__

    def __repr__(self):
        return "{...} (%d entries)" % len(self)

sys_modules = opaque_dict(sys.modules)

def evaluate_expression(pragma, expr):
    key = "%s(%s)" % (pragma, expr)
    try:
        symbol_mapping, parts, source = _expr_cache[key]
    except KeyError:
        translator = component.getUtility(IExpressionTranslator, name=pragma)
        parts = translator.tales(expr)
        stream = generation.CodeIO(symbols=config.SYMBOLS)
        assign = clauses.Assign(parts, 'result')
        assign.begin(stream)
        assign.end(stream)
        source = stream.getvalue()

        symbol_mapping = parts.symbol_mapping.copy()
        if isinstance(parts, types.parts):
            for value in parts:
                symbol_mapping.update(value.symbol_mapping)    

        _expr_cache[key] = symbol_mapping, parts, source

    # acquire template locals and update with symbol mapping
    frame = sys._getframe()
    while frame.f_locals.get('econtext', _marker) is _marker:
        frame = frame.f_back
        if frame is None:
            raise RuntimeError, "Can't locate template frame."

    _locals = frame.f_locals
    _locals.update(symbol_mapping)
    _locals.update(_locals['econtext'])

    # to support dynamic scoping (like in templates), we must
    # transform the code to take the scope locals into account; for
    # efficiency, this is cached for reuse
    code_cache_key = key, tuple(_locals)

    try:
        code = _expr_cache[code_cache_key]
    except KeyError:
        suite = codegen.Suite(source, _locals)
        code = compiler.compile(
            suite.source, 'dynamic_path_expression.py', 'exec')
        _expr_cache[code_cache_key] = code

    # execute code and return evaluation
    _locals[config.SYMBOLS.lookup_attr] = codegen.lookup_attr
    exec code in _locals
    return _locals['result']

def evaluate_path(expr):
    return evaluate_expression('path', expr)

def evaluate_exists(expr):
    try:
        return evaluate_expression('exists', expr)
    except NameError:
        return False

class BaseTemplate(template.PageTemplate):
    content_type = None
    default_parser = language.Parser()
    version = 2

    def bind(self, ob=None, request=None, macro=None, global_scope=True):
        def render(target_language=None, request=request, **kwargs):
            context = self._pt_get_context(ob, request, kwargs)
            request = request or context.get('request')
            if target_language is None:
                target_language = request.get("LANGUAGE", None)
                if target_language is None:
                    try:
                        target_language = i18n.negotiate(request)
                    except:
                        target_language = None

            context['target_language'] = target_language
            context['econtext'] = utils.econtext(context)

            # bind translation-method to request
            def translate(
                msgid, domain=None, mapping=None, target_language=None, default=None):
                return fast_translate(
                    msgid, domain, mapping, request, target_language, default)
            context[config.SYMBOLS.translate] = translate

            if request is not None and not isinstance(request, basestring):
                content_type = self.content_type or 'text/html'
                response = request.response
                if response and not response.getHeader("Content-Type"):
                    response.setHeader(
                        "Content-Type", content_type)

            if macro is None:
                return self.render(**context)
            else:
                return self.render_macro(
                    macro, global_scope=global_scope, parameters=context)

        return BoundPageTemplate(self, render)

    def __call__(self, *args, **kwargs):
        bound_pt = self.bind()
        return bound_pt(*args, **kwargs)

    def _pt_get_context(self, instance, request, kwargs):
        return dict(
            options=kwargs,
            request=request,
            path=evaluate_path,
            exists=evaluate_exists,
            nothing=None,
            modules=sys_modules)

class BaseTemplateFile(BaseTemplate, template.PageTemplateFile):
    """If ``filename`` is a relative path, the module path of the
    class where the instance is used to get an absolute path."""

    def __init__(self, filename, path=None, content_type=None, **kwargs):
        if path is not None:
            filename = os.path.join(path, filename)

        if not os.path.isabs(filename):
            for depth in (1, 2):
                frame = sys._getframe(depth)
                package_name = frame.f_globals.get('__name__', None)
                if package_name is not None and package_name != self.__module__:
                    module = sys.modules[package_name]
                    try:
                        path = module.__path__[0]
                    except AttributeError:
                        path = module.__file__
                        path = path[:path.rfind(os.sep)]
                    break
                else:
                    package_path = frame.f_globals.get('__file__', None)
                    if package_path is not None:
                        path = os.path.dirname(package_path)
                        break

            if path is not None:
                filename = os.path.join(path, filename)

        template.PageTemplateFile.__init__(
            self, filename, **kwargs)

        # Set content-type last, so that we can override whatever was
        # magically sniffed from the source template.
        self.content_type = content_type

class PageTemplate(BaseTemplate):
    """Page Templates using TAL, TALES, and METAL.

    This class is suitable for standalone use or class
    property. Keyword-arguments are passed into the template as-is.

    Initialize with a template string."""

    version = 1

    def __get__(self, instance, type):
        if instance is not None:
            return self.bind(instance)
        return self

class PageTemplateFile(BaseTemplateFile, PageTemplate):
    """Page Templates using TAL, TALES, and METAL.

    This class is suitable for standalone use or class
    property. Keyword-arguments are passed into the template as-is.

    Initialize with a filename."""

class ViewPageTemplate(PageTemplate):
    """Template class suitable for use with a Zope browser view; the
    variables ``view``, ``context`` and ``request`` variables are
    brought in to the local scope of the template automatically, while
    keyword arguments are passed in through the ``options``
    dictionary. Note that the default expression type for this class
    is 'path' (standard Zope traversal)."""

    def _pt_get_context(self, view, request, kwargs):
        context = kwargs.get('context')
        if context is None:
            context = view.context
        request = request or kwargs.get('request') or view.request
        return dict(
            view=view,
            context=context,
            request=request,
            path=evaluate_path,
            exists=evaluate_exists,
            options=kwargs,
            nothing=None,
            modules=sys_modules)

    def __call__(self, _ob=None, context=None, request=None, **kwargs):
        kwargs.setdefault('context', context)
        kwargs.setdefault('request', request)
        bound_pt = self.bind(_ob)
        return bound_pt(**kwargs)

class ViewPageTemplateFile(ViewPageTemplate, PageTemplateFile):
    """If ``filename`` is a relative path, the module path of the
    class where the instance is used to get an absolute path."""

class BoundPageTemplate(object):
    """When a page template class is used as a property, it's bound to
    the class instance on access, which is implemented using this
    helper class."""

    def __init__(self, pt, render):
        object.__setattr__(self, 'im_self', pt)
        object.__setattr__(self, 'im_func', render)

    macros = property(lambda self: self.im_self.macros)
    filename = property(lambda self: self.im_self.filename)

    def __call__(self, *args, **kw):
        kw.setdefault('args', args)
        return self.im_func(**kw)

    def __setattr__(self, name, v):
        raise AttributeError("Can't set attribute", name)

    def __repr__(self):
        return "<%s.Bound%s %r>" % (
            type(self.im_self).__module__,
            type(self.im_self).__name__, self.filename)



import unittest

class LoadTests:
    def _makeOne(self, search_path=None, auto_reload=False, cachedir=None):
        klass = self._getTargetClass()
        return klass(search_path, auto_reload, cachedir)

    def _getTargetClass(self):
        from z3c.pt.loader import TemplateLoader
        return TemplateLoader

    def test_load_relative(self):
        import os
        here = os.path.dirname(__file__)
        loader = self._makeOne(search_path = [here])
        
        result = self._load(loader, 'view.pt')
        self.assertEqual(result.filename, os.path.join(here, 'view.pt'))

    def test_consecutive_loads(self):
        import os
        here = os.path.dirname(__file__)
        loader = self._makeOne(search_path = [here])
        
        self.assertTrue(
            self._load(loader, 'view.pt') is self._load(loader, 'view.pt'))

    def test_load_relative_badpath_in_searchpath(self):
        import os
        here = os.path.dirname(__file__)
        loader = self._makeOne(search_path = [os.path.join(here, 'none'), here])
        result = self._load(loader, 'view.pt')
        self.assertEqual(result.filename, os.path.join(here, 'view.pt'))

    def test_load_abs(self):
        import os
        here = os.path.dirname(__file__)
        loader = self._makeOne()
        abs = os.path.join(here, 'view.pt')
        result = self._load(loader, abs)
        self.assertEqual(result.filename, abs)

class LoadPageTests(unittest.TestCase, LoadTests):
    def _load(self, loader, filename):
        return loader.load_page(filename)

class LoadTextTests(unittest.TestCase, LoadTests):
    def _load(self, loader, filename):
        return loader.load_text(filename)
    
def test_suite():
    import sys
    return unittest.findTestCases(sys.modules[__name__])



# This test is adopted from the spifire project, which has an extensive
# test suite of different templating languages.

# You can run it via: bin/py benchmark/benchmark/bigtable.py

import sys
import timeit

from zope.pagetemplate import pagetemplate

from chameleon.zpt import template

bigtable_z3c = template.PageTemplate("""
<table xmlns="http://www.w3.org/1999/xhtml"
xmlns:tal="http://xml.zope.org/namespaces/tal">
<tr tal:repeat="row table">
<td tal:repeat="column row.values()" tal:content="column">
</td></tr></table>
""")

bigtable_zope = pagetemplate.PageTemplate()
bigtable_zope.pt_edit("""\
<table xmlns="http://www.w3.org/1999/xhtml"
xmlns:tal="http://xml.zope.org/namespaces/tal">
<tr tal:repeat="row python: options['table']">
<td tal:repeat="c python: row.values()" tal:content="c">
</td></tr></table>""", 'text/xhtml')


table = [dict(a=1,b=2,c=3,d=4,e=5,f=6,g=7,h=8,i=9,j=10) \
         for x in range(1000)]

def setup():
    import z3c.pt
    from zope.component.testing import setUp
    from zope.configuration import xmlconfig

    setUp()
    xmlconfig.XMLConfig('configure.zcml', z3c.pt)()

def test_z3c():
    """z3c.pt template"""
    data = bigtable_z3c(table=table)

def test_zope():
    """zope.pagetemplate template"""
    data = bigtable_zope(table=table)

def run(which=None, number=10):
    tests = ['test_z3c', 'test_zope']

    if which:
        tests = ['test_z3c']

    for test in tests:
        setup = 'from __main__ import setup, %s; setup(); %s()' % (test, test)

        t = timeit.Timer(setup=setup,
                         stmt='%s()' % test)
        time = t.timeit(number=number) / number

        if time < 0.00001:
            result = '   (not installed?)'
        else:
            result = '%16.2f ms' % (1000 * time)
        print '%-35s %s' % (getattr(sys.modules[__name__], test).__doc__, result)


if __name__ == '__main__':
    which = None
    if '-p' in sys.argv:
        which = True
        from cProfile import Profile
        import pstats
        profiler = Profile()
        setup()
        test_z3c()
        profiler.runcall(test_z3c)
        profiler.dump_stats('template.prof')
        stats = pstats.Stats(profiler)
        stats.strip_dirs()
        stats.sort_stats('time', 'calls')
        stats.print_stats()
    else:
        run(which, 1)
